use crate::content;
use cgmath::{Deg, Matrix2, Point2, Point3, Rad, Vector2};
use rand::Rng;

/// Instructions to create a new particle
pub struct ParticleBuilder {
	/// The sprite to use for this particle
	pub sprite: content::SpriteHandle,

	/// This object's center, in world coordinates.
	pub pos: Point2<f32>,

	/// This particle's velocity, in world coordinates
	pub velocity: Vector2<f32>,

	/// This particle's angle
	pub angle: Rad<f32>,

	/// This particle's angular velocity (rad/sec)
	pub angvel: Rad<f32>,

	/// This particle's lifetime, in seconds
	pub lifetime: f32,

	/// The size of this particle,
	/// given as height in world units.
	pub size: f32,

	/// Fade this particle out over this many seconds as it expires
	pub fade: f32,
}

impl ParticleBuilder {
	/// Create a ParticleBuilder from an Effect
	pub fn from_content(
		effect: &content::Effect,
		pos: Point2<f32>,
		parent_angle: Rad<f32>,
		parent_velocity: Vector2<f32>,
		target_velocity: Vector2<f32>,
	) -> Self {
		let mut rng = rand::thread_rng();

		let velocity = {
			let a =
				rng.gen_range(-effect.velocity_scale_parent_rng..=effect.velocity_scale_parent_rng);
			let b =
				rng.gen_range(-effect.velocity_scale_target_rng..=effect.velocity_scale_target_rng);

			let velocity = ((effect.velocity_scale_parent + a) * parent_velocity)
				+ ((effect.velocity_scale_target + b) * target_velocity);

			Matrix2::from_angle(Rad(
				rng.gen_range(-effect.direction_rng..=effect.direction_rng)
			)) * velocity
		};

		// Rad has odd behavior when its angle is zero, so we need extra checks here
		let angvel = if effect.angvel_rng == 0.0 {
			effect.angvel
		} else {
			Rad(effect.angvel.0 + rng.gen_range(-effect.angvel_rng..=effect.angvel_rng))
		};
		let angle = if effect.angle_rng == 0.0 {
			parent_angle + effect.angle
		} else {
			parent_angle + effect.angle + Rad(rng.gen_range(-effect.angle_rng..=effect.angle_rng))
		};

		ParticleBuilder {
			sprite: effect.sprite,
			pos,
			velocity,

			angle,
			angvel,

			lifetime: 0f32
				.max(effect.lifetime + rng.gen_range(-effect.lifetime_rng..=effect.lifetime_rng)),

			// Make sure size isn't negative. This check should be on EVERY rng!
			size: 0f32.max(effect.size + rng.gen_range(-effect.size_rng..=effect.size_rng)),

			fade: 0f32.max(effect.fade + rng.gen_range(-effect.fade_rng..=effect.fade_rng)),
		}
	}
}

/// The location of a UI element, in one of a few
/// possible coordinate systems.
///
/// Positive Y always points up,
/// positive X always points right.
#[derive(Debug, Clone)]
pub enum AnchoredUiPosition {
	/// Position of this sprite's center,
	/// relative to the nw corner of the window.
	NwC(Point2<f32>),

	/// Position of this sprite's nw corner,
	/// relative to the nw corner of the window.
	NwNw(Point2<f32>),

	/// Position of this sprite's ne corner,
	/// relative to the nw corner of the window.
	NwNe(Point2<f32>),

	/// Position of this sprite's sw corner,
	/// relative to the nw corner of the window.
	NwSw(Point2<f32>),

	/// Position of this sprite's se corner,
	/// relative to the nw corner of the window.
	NwSe(Point2<f32>),
}

/// A sprite that represents a ui element
#[derive(Debug, Clone)]
pub struct UiSprite {
	/// The sprite to draw
	pub sprite: content::SpriteHandle,

	/// This object's position, in logical (dpi-adjusted) pixels
	pub pos: AnchoredUiPosition,

	/// This sprite's color will be multiplied by this value.
	/// If this is None, color will not be changed.
	pub color: Option<[f32; 4]>,

	/// The size of this sprite, in logical (dpi-adjusted) pixels
	pub dimensions: Point2<f32>,

	/// This sprite's rotation, measured ccw
	pub angle: Deg<f32>,
}

/// A sprite that represents a world object:
/// Ships, planets, debris, etc
#[derive(Debug, Clone)]
pub struct ObjectSprite {
	/// The sprite to draw
	pub sprite: content::SpriteHandle,

	/// This object's center, in world coordinates.
	pub pos: Point3<f32>,

	/// The size of this sprite,
	/// given as height in world units.
	pub size: f32,

	/// This sprite's rotation
	/// (relative to north, measured ccw)
	/// Note that this is different from the angle used by our physics system.
	pub angle: Deg<f32>,

	/// Sprites that should be drawn relative to this sprite.
	pub children: Option<Vec<ObjectSubSprite>>,
}

/// A sprite that is drawn relative to an ObjectSprite.
#[derive(Debug, Clone)]
pub struct ObjectSubSprite {
	/// The sprite to draw
	pub sprite: content::SpriteHandle,

	/// This object's position, in world coordinates.
	/// This is relative to this sprite's parent.
	pub pos: Point3<f32>,

	/// The size of this sprite,
	/// given as height in world units.
	pub size: f32,

	/// This sprite's rotation
	/// (relative to north, measured ccw)
	/// Just as position, this is relative to this
	/// subsprite's parent sprite.
	pub angle: Deg<f32>,
}
