use cgmath::{Deg, InnerSpace, Point3, Vector2};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyHandle},
	geometry::ColliderHandle,
};

use crate::util;
use galactica_gameobject as object;
use galactica_render::ObjectSprite;

/// A single projectile in the world
#[derive(Debug)]
pub struct ProjectileWorldObject {
	/// This projectile's game data
	pub projectile: object::Projectile,

	/// This projectile's rigidbody
	pub rigid_body: RigidBodyHandle,

	/// This projectile's collider
	pub collider: ColliderHandle,

	/// This projectile's size variation
	pub size_rng: f32,
}

impl ProjectileWorldObject {
	/// Make a new projectile
	pub fn new(
		projectile: object::Projectile,
		rigid_body: RigidBodyHandle,
		collider: ColliderHandle,
	) -> Self {
		let mut rng = rand::thread_rng();
		let size_rng = projectile.content.size_rng;
		ProjectileWorldObject {
			rigid_body,
			collider,
			projectile,
			size_rng: rng.gen_range(-size_rng..=size_rng),
		}
	}

	/// Get this projectiles' sprite
	pub fn get_sprite(&self, r: &RigidBody) -> ObjectSprite {
		let pos = util::rigidbody_position(r);
		let rot = util::rigidbody_rotation(r);

		// Sprites point north at 0 degrees
		let ang: Deg<f32> = rot.angle(Vector2 { x: 1.0, y: 0.0 }).into();

		ObjectSprite {
			sprite: self.projectile.content.sprite,
			pos: Point3 {
				x: pos.x,
				y: pos.y,
				z: 1.0,
			},
			size: 0f32.max(self.projectile.content.size + self.size_rng),
			angle: -ang,
			children: None,
		}
	}
}
